
%{
#include "yacc2html.h"

extern char const * program_name, * input_file,
	* basis, * yacc_sed_file, * yacc_file,
	* lex_sed_file, * lex_file, * title, * fmt_terminal_url,
	* fmt_nonterminal_name;

extern int recursive_links, union_links, token_terminals,
	ignore_terminals, c_code_suppression;

/* These format strings are used with the print() function in output.c.
 *
 * %<xTEXT%>x sets TEXT in
 *	x = i	italic
 *	  = b	bold 
 *	  = e	emphasized
 *	  = t	type-written
 *	  = c	code-like font.
 *
 * In addition, %<m and %>m delimit areas of markup than may be enclosed
 * neither by font attributes nor by anchors; 
 *
 * %<nTEXT%>n  makes TEXT an anchor with a name that is passed as 
 *  	a varargs char * argument;
 *
 * %<hTEXT%>h  makes TEXT a link source with a href that is passed as 
 *  	a varargs char * argument;
 *
 * %<HTEXT%>H  makes TEXT a local link source with a href that is passed as 
 *  	a varargs char * argument (that is, H equals h plus a free '#')
 *
 * %_ 	typesets a sentence break (wide space).  
 *
 * Anything that is not a % command is copied to the standard output.
 * All output that is not markup is entity-quoted; for example, a '<' that
 * is not explicitly transmitted as Markup (e.g. using %<m .. %>m) will
 * be turned into  &lt; .
 */

/*  fmt_header is printed at the beginning of the run, with a single
 *  string argument that is the (predefined or user-specified) title
 *  of the document.
 */
#define		fmt_header	\
	"%<m<html><head>\n<title>%>m%s%<m</title>\n</head>\n<body><pre>%>m\n"

/*  fmt_footer is printed at the end of the run, with no arguments.
 */
#define		fmt_footer	"%<m</pre>\n</body>\n</html>%>m\n"

/*  fmt_syntax is used for syntactic elements of YACC, that is,  YACC's
 *  colon :, bar |, and semicolon ;, as well as left and right curls,
 *  the section separators, and the angle brackets around type tags.
 *  To print all syntactic elements bold, use "%<b%s%>b".
 */
#define		fmt_syntax	"%s"

/*  FMT_CODE is used for C code, string and character literals, and the
 *  %union contents.
 */
#define		fmt_code	"%s"

/*  FMT_TAG is used for the type tag (which must be the name of a
 *  member of the %union structure) that specifies which variant of
 *  the %union a token implements.
 */
#define		fmt_tag		"%s"

/*  FMT_NTDEF is used for definitions of nonterminals, that is,
 *  the identifier before the ':'.  To make the nonterminals appear
 *  in italic font, use "%<i%s%>i".
 */
#define		fmt_ntdef	"%s"

/*  FMT_NTUSE is used for references to nonterminals that are
 *  defined elsewhere.
 */
#define		fmt_ntuse	"%s"

/*  FMT_TERMINAL is used for terminal symbols, that is, symbols that
 *  are declared using %token and that are most likely defined
 *  in a lex file somewhere else.
 */
#define		fmt_terminal	"%s"

/*  These hashtables help keep track of which identifiers have been
 *  declared as terminals (if they are, they are in the "terminals"
 *  hashtable), and which have been defined as nonterminals.
 *
 *  In YACC, multiple redefinitions of the same nonterminal are
 *  legal---the rules just add up---but in HTML, we should have only
 *  one anchor for a nonterminal.  If a symbol is in the "nonterminals"
 *  hashtable, we think of it as anchored.
 */
Hashtable terminals, nonterminals, defined_symbols;

/* The nonterminal whose production the parser is reading right now.
 * Used to find out which nonterminal uses are recursive and which aren't.
 */
struct list {				/* The yacc file is compiled into */
	int		type;		/* a linked list of these tokens, */
	char const      * text;		/* and then massaged into HTML.	  */
	struct list	* next;
};					/* possible token types: 	*/

#define	T_SYNTAX	1		/* lex syntactic element	*/
#define	T_CODE		2		/* c code			*/
#define	T_UCODE		3		/* %union c code		*/
#define	T_LIT		4		/* literals			*/
#define	T_WS		5		/* white space, incl. comments	*/
#define	T_IDENT		6		/* identifier			*/
#define T_NDEF		7		/* yacc nonterminal definition	*/
#define T_NREDEF	8		/* yacc nonterminal redefinition */
#define T_NUSE		9		/* yacc nonterminal use		*/
#define T_NREC		10		/* yacc nonterminal use, recursive */
#define T_TDEF		11		/* lex definition, i.e. %token	*/
#define T_TUSE		12		/* lex token use		*/
#define T_TAG		13		/* type tag			*/

%}

/*  I am extending yacc's grammar in two aspects: (1), string literals
 *  are legal where character literals are legal, and (2), the first
 *  ("declarations") part is as optional as the trailer.  Both changes
 *  are intended to allow conversion of arbitrary grammars, rather than
 *  just YACC's specific input format.
 */
%token	<list> WS CCODE PREFIX ID CURLS TAG NUMBER RWORD TOKEN
%token	 MARK START UNION PREC
%type	<list> ws spec tail defs def nlist name rules rule body prec elem elems 
%type	<list> r_rules

%union {
	struct list * list;
}

%%
spec:	  ws defs MARK ws rules tail  {
			  $$ = f("llslll", $1, $2, "%%", $4, $5, $6);
/*		   dump("$$ in fullspec", $$); 	*/
			  process($$);
			}
spec:	ws rules tail { $$ = f("lll", $1, $2, $3);
/*			   dump("$$ in spec", $$); */
			  process($$);
			}
	;

ws:	  ws WS			{ $$ = f("ll", $1, $2);			}
	| /* empty */		{ $$ = f(""); 				}
	;

tail:	  MARK 			{ buf_n = 0;
				  skipahead(EOF, EOF, "trailer");
				  $$ = f("sc", "%%", buf_s);		}
	| ws
	;

defs: 	  defs def		{ $$ = f("ll", $1, $2);
				/* dump("defs", $$ ); */}
	| error  WS		{
	yyerrok; 	parser_error("syntax error in definition"); }
	| ws
	;

def: 	  START ws ID ws	{ $$ = f("slll", "%start", $2, $3, $4);
/* $3->type = T_IDENT; */
/* dump("startdef", $$); */
				}
	| UNION ws CCODE ws	{ struct list * l;
				  for (l = $3; l; l = l->next)
					if (l->type == T_CODE) {
						l->type = T_UCODE;
						break;
					}
				  $$ = f("slll", "%union", $2, $3, $4);
				}
	| TOKEN ws TAG ws nlist { struct list * l;
				  for (l = $5; l; l = l->next)
					if (l->type == T_IDENT) def_terminal(l);
				  $$ = f("lllll", $1, $2, $3, $4, $5);
				}
	| TOKEN ws nlist	{ struct list * l;
				  for (l = $3; l; l = l->next)
					if (l->type == T_IDENT) def_terminal(l);
				  $$ = f("lll", $1, $2, $3);	
				}
	| RWORD ws TAG ws nlist { struct list * l;
				  for (l = $5; l; l = l->next)
					if (l->type == T_IDENT) {
						if (terminalP(l))
							use_terminal(l);
						else use_nonterminal(l);
					}
				  $$ = f("lllll", $1, $2, $3, $4, $5);
				}
	| RWORD ws nlist	{ struct list * l;
				  for (l = $3; l; l = l->next) {
					if (l->type == T_IDENT) {
						if (terminalP(l))
							use_terminal(l);
						else use_nonterminal(l);
					}
				  }
				  $$ = f("lll", $1, $2, $3);	
				}
	| CURLS ws		{ $$ = f("ll", $1, $2); 		}
	;

nlist:	  name
	| nlist name 		{ $$ = f("ll", $1, $2);			}
	| nlist ',' ws name 	{ $$ = f("lsll", $1, ",", $3, $4); 	}
	;

name:	  ID ws			{ $$ = f("ll", $1, $2);			}
	| ID ws NUMBER ws	{ $$ = f("llll", $1, $2, $3, $4);	}
	;

rules:	  PREFIX ws body prec r_rules {
			$$ = f("lllll", $1, $2, $3, $4, $5);
		}
	;

r_rules:  /* empty */ 		{ $$ = f("");				}
	| r_rules rule		{ $$ = f("ll",  $1, $2);		}
	;

rule:	  PREFIX ws body prec 	{ $$ = f("llll", $1, $2, $3, $4); 	}
	| '|' ws body prec	{ $$ = f("slll", "|", $2, $3, $4);	}
	;

body:	  /* empty */		{ $$ = f("");				}
	| elems
	;

elems:	  elem			
	| elems elem		{ $$ = f("ll", $1, $2); 		}
	;

elem:	  ID ws			{ $$ = f("ll", $1, $2);
/*
				  if ($1->type == T_IDENT)
					if (terminalP($1)) use_terminal($1);
				  	else use_nonterminal($1);
 */
				}
	| CCODE ws		{ $$ = f("ll", $1, $2);			}
	;

prec:	  /* empty */		{ $$ = f("");				}
	| PREC ws ID ws		{ $$ = f("slll", "%prec", $2, $3, $4);
				  if ($3->type == T_IDENT)
					  if (terminalP($3)) use_terminal($3);
					  else use_nonterminal($3);
				} 
	| PREC ws ID ws CCODE ws  { struct list * l;
				  $$ = f("slllll", "%prec", $2, $3, $4, $5, $6);
				  if ($3->type == T_IDENT)
					  if (terminalP($3)) use_terminal($3);
					  else use_nonterminal($3);
				}
	| prec ';' ws		{ $$ = f("lsl", $1, ";", $3);		}
	;

%%
#include	"c.lex.c"

yyerror(s) char * s;
{
	parser_error("Parser error: %s", s);
	return 0;
}

char *
Realloc(char * ptr, size_t size)
{
	char	* c;
	if (size) {
		c = (char *)(ptr ? realloc(ptr, size) : malloc(size));
		if (!c) parser_error(
			"failed to malloc/realloc %lx to %d bytes\n",
			ptr, size);
		return c;
	}
	else {
		if (ptr) free(ptr);
		return (char *)0;
	}
}

use_terminal(struct list * l)
{
	assert (l->type == T_IDENT);
	assert (terminalP(l));
	l->type = T_TUSE;
}

def_terminal(struct list * l)
{
	assert (l->type == T_IDENT);
	l->type = htst(&terminals, l->text) ? T_TUSE : T_TDEF;
}

terminalP(struct list * l)
{
	return l->type == T_IDENT && terminals && hash(terminals, l->text, 0);
}

def_nonterminal(struct list * l)
{
	assert (l->type == T_IDENT);

	if (htst(&nonterminals, l->text)) {
		l->type = T_NREDEF;
	} else {
		l->type = T_NDEF;
	}
}

recuse_nonterminal(struct list * l)
{
	assert (l->type == T_IDENT);
	l->type = T_NREC;
}

use_nonterminal(struct list * l)
{
	assert (l->type == T_IDENT);
	l->type = T_NUSE;
}

struct list * tstring(char const * str, int tp)
{
	struct list * s;

	s = TMALLOC(struct list, 1);
	assert(s);	/* TMALLOC error-handles */

	s->type  = tp;
	s->text  = str ? intern(str) : 0;
	s->next  = (struct list *)0;

	return s;
}

struct list * f(char const * fmt, ...)
{
	char const  * s;
	struct list * head = (struct list *)0,
		   ** hp = &head;
	int	      type;
	va_list	      ap;

	va_start (ap, fmt);
	for (; *fmt; fmt++) switch (*fmt) {
	case 'l': *hp = va_arg(ap, struct list *);
/*		  dump("f() list argument",  *hp);
		  dump("f() whole list",  head);
 */
		  while (*hp) hp = &(*hp)->next;
		  break;

	case 's': type = T_SYNTAX; goto single_string;
	case 'w': type = T_WS;     goto single_string;
	case 'i': type = T_IDENT;  goto single_string;
	case 'd': type = T_NDEF;   goto single_string;
	case 'u': type = T_NUSE;   goto single_string;
	case 'D': type = T_TDEF;   goto single_string;
	case 'U': type = T_TUSE;   goto single_string;
	case 'c': type = T_CODE;   goto single_string;
	case '\'': type = T_LIT;  goto single_string;
	case 'T': type = T_TAG;
 single_string: s = va_arg(ap, char const *); assert(s);
		*hp = tstring(s, type);
/*		dump("f() string argument",  *hp);
		dump("f() whole list",  head);
*/
		hp = &(*hp)->next;
		break;

	default: fprintf(stderr, "%s: unknown type specifier '%c' (%d)\n",
			program_name, *fmt, *fmt);
		exit(1);
	}

	if (!head) head = tstring("", T_WS);
/*	dump("f() result",  head);	*/
	return head;
}

char const * uprint (char const * fmt, char const * s)
{
	static char * 	url_s;		/* never freed, reused	*/
	static size_t 	url_m;		/* # bytes for url_s	*/

	size_t 	 	 n = strlen(s) + strlen(fmt) + 1;
	char const      *f, *q;
	char		*u;

	if (n > url_m) { url_s = TREALLOC(char, url_s, n); url_m = n; }

	for (f = fmt, u = url_s; *f; f++) {
		if (*f != '%') *u++ = *f;
		else switch (*++f) {
		case 'l':
			n += strlen(lex_file); *u = 0;
			if (n > url_m) url_s = TREALLOC(char, url_s, url_m = n);
			for (u = url_s+strlen(url_s), q = lex_file;
				*u = *q++; u++) ;
			break;
		case 'y':
			n += strlen(yacc_file); *u = 0;
			if (n > url_m) url_s = TREALLOC(char, url_s, url_m = n);
			for (u = url_s+strlen(url_s), q = yacc_file;
				*u = *q++; u++) ;
			break;
		case 's':
			n += strlen(s); *u = 0;
			if (n > url_m) url_s = TREALLOC(char, url_s, url_m = n);
			for (u = url_s+strlen(url_s), q = s; *u = *q++; u++) ;
			break;
		case 'b':
			n += strlen(basis); *u = 0;
			if (n > url_m) url_s = TREALLOC(char, url_s, url_m = n);
			for (u=url_s+strlen(url_s), q=basis; *u = *q++; u++) ;
			break;

		case '%': *u++ = '%'; break;

		default: fprintf(stderr,
	"%s: unknown format specification character %c (%x) in \"%s\"\n",
				program_name, *f, fmt);
			exit(1);
		}
	}
	*u = 0;
	return url_s;
}

char const * quote_url(char const * c)
{
	static char * 	url_s;		/* never freed, reused	*/
	static size_t 	url_m;		/* # bytes for url_s	*/

	size_t 		n = strlen(c) * 4 + 1;	/* 4 for %255 	*/
	char          * p;

	if (n > url_m) {
		url_s = TREALLOC(char, url_s, n);
		url_m = n;
	}

	for (p = url_s; *c; c++) {
		if ( isalnum(*c)
		   || *c == '$' || *c == '-' || *c == '_' || *c == '@'
		   || *c == '.' || *c == '&' || *c == '+' || *c == '-')
			*p++ = *c;
		else {
			sprintf(p, "%%%2.2x", *(unsigned char *)c);
			p += strlen(p);
		}
	}
	*p = 0;
	return url_s;
}


char const * nonterminal_name (char const * s)
{
	return intern(uprint(fmt_nonterminal_name, quote_url(s)));
}

char const * terminal_url (char const * s)
{
	char const * p;
	p = intern(uprint(fmt_terminal_url, quote_url(s)));
	return p;
}

static void print_link (
	int c,
	char const * mid,
	char const * url,
	char const * text)
{
	static char pre[4] = "%<x", post[4] = "%>x";

	assert(mid);

	pre[2] = post[2] = c;

	if (c) {
		if (url) print(pre, url);
		else print(pre);
	}
	print(mid, text);
	if (c) print (post);
}

void define_link(char const * src, char const * target)
{
	extern Hashtable defined_symbols;
	char const ** p;

	if (!defined_symbols) defined_symbols = htable(char const *, 20);
	if (p = haccess(defined_symbols, char const *, src)) {
		fprintf(stderr, "%s: redefinition of \"%s\" (replaced)\n",
			program_name, src);
	}
	else p = hnew(defined_symbols, char const *, src);
	assert(p);
	*p = target;	/* points to *argv, isn't modified. */
}

char const ** definedP (char const * src)
{
	char const ** p;

	if (!defined_symbols) return 0;
	return haccess(defined_symbols, char const *, src);
}

static void print_def (char const ** p, char const * fmt)
{
	char const * n;

	assert(p); assert(fmt);

	n = hname(defined_symbols, char const *, p);
	assert(n);

	if (*p) print("%<h%s%>h", *p, n);
	else print(fmt, n);
}


process (struct list * l)
{
	struct list * k;
	char const ** p;
	char const  * this_nonterminal = 0;

	/*  Mark recursive nonterminal uses as such. 
	 */
	for (k = l; k; k = k->next) {
		switch (k->type) {
		case T_NDEF: 	this_nonterminal = k->text; 	break;
		case T_NREDEF:	this_nonterminal = 0;		break;
		case T_IDENT:
			if (terminalP(k)) use_terminal(k);
			else k->type = ( k->text && k->text == this_nonterminal
				       ? T_NREC
				       : T_NUSE);
		default:
			break;
		}
	}

	print(fmt_header, title ? title : input_file);

	for (; l; l = l->next) {

		if (!l->text || !*l->text) continue;
again:		switch (l->type) {
		case T_SYNTAX:
			if (union_links && !strcmp(l->text, "%union")) 
				print_link('n', fmt_syntax, "%63union",
					l->text);
			else if (  c_code_suppression
				&& !strcmp(l->text, "%{")) {
				while (l && strcmp(l->text, "%}")) l = l->next;
				while (l && l->type == T_WS) l = l->next;
				break;
			}
			else if (p = definedP(l->text)) print_def(p,fmt_syntax);
			else print(fmt_syntax, l->text);
			break;

		case T_LIT:
		case T_UCODE:
			if (p = definedP(l->text)) print_def(p, fmt_code);
			print(fmt_code, l->text); break;

		case T_CODE:
			if (!c_code_suppression) {
				if (p = definedP(l->text))
					print_def(p, fmt_code);
				else print(fmt_code, l->text);
			}
			else if (l->next && l->next->type == T_WS)
				l = l->next;	/* skip next ws, too */
			break;


		case T_WS: print("%s", l->text); break;

		case T_IDENT:
			if (terminalP(l)) use_terminal(l);
			else use_nonterminal(l);
			assert(l->type != T_IDENT);
			goto again;

		/*
		 *  Nonterminals.  Four possible cases:
		 * 
		 *  - definition;
		 *  - redefinition; looks like definition,
		 *    but can't have name tag.
		 *  - use; linked to definition
		 *  - recursive rule, i.e. in the body of the
		 *    rule that has been tagged.  Looks like
		 *    use, but might not be linked to parent.
		 */
		case T_NDEF:
			if (p = definedP(l->text)) print_def(p, fmt_ntdef);
			else print_link('n', fmt_ntdef,
				nonterminal_name(l->text), l->text);
			break;

		case T_NREDEF:
			if (p = definedP(l->text)) print_def(p, fmt_ntdef);
			else print_link('H', fmt_ntdef, 
				nonterminal_name(l->text), l->text);
			break;

		case T_NREC:
			if (p = definedP(l->text)) {
				print_def(p, fmt_ntuse);
				break;
			}
			if (!recursive_links) {
				print(fmt_ntuse, l->text);
				break;
			}
			/* else fall through */

		case T_NUSE:
			if (p = definedP(l->text)) print_def(p, fmt_ntuse);
			else if (!hash(nonterminals, l->text, 0)) {
				print(fmt_ntuse, l->text);
			}
			else print_link('H', fmt_ntdef,
				nonterminal_name(l->text), l->text);
			break;

		/* 
		 *  Terminals.  Three options:
		 *  - link to terminal declaration at top of local
		 *    grammar ("local")
		 *  - link to external file  (the default) 
		 *  - don't link at all.  ("ignore")
		 */
		case T_TDEF:
		case T_TUSE:
			if (p = definedP(l->text))
				print_def(p, fmt_terminal);

			else if (ignore_terminals
				|| !hash(terminals, l->text, 0))
				print(fmt_terminal, l->text);

			else if (token_terminals) print_link(
				"Hn"[l->type == T_TDEF], fmt_terminal,
					terminal_url(l->text), l->text); 

			else print_link('h', fmt_terminal,
					terminal_url(l->text), l->text);

			break;

		/* Tags.  Link to #%union if the user specified -u .
		 */
		case T_TAG:
			if (p = definedP(l->text)) print_def(p, fmt_tag);
			else print_link("H"[!union_links], fmt_tag,
				"%63union", l->text);
			break;

		default:
			fprintf(stderr, "unexpected list type %d\n", l->type);
			return -1;
		}
	}
	print(fmt_footer);

	return 0;
}
