
#include	"yacc2html.h"

typedef struct slot {

	/* application data
	 */

	char		* next;		/* collision chain	*/
	unsigned long	  hash;		/* hash value of name	*/

	/* '\0'-terminated string name
	 */

} SLOT, * Slot;

#define	INFO(   h, base )	((Slot)((base) + (h)->o))
#define NEXT(   h, base )	(INFO(h, base)->next)
#define STRING( h, base )	((base) + (h)->s)

static unsigned long hashf(char const * s)
{
	register unsigned long i = 0;
	while (*s) i = (i << 3) + (unsigned char)*s++;
	return i;
}

static unsigned long round(unsigned long size)
{
	unsigned long i;

	if (size >= 1) size += size - 1;
	for (i = 1; i && i < size; i <<= 1) ;

	return i;
}

Hashtable hashtable(unsigned int elsize, unsigned int m)
{
	Hashtable t;
	char   ** c, ** e;

	t       = TMALLOC( HASHTABLE, 1 );

	t->o    = elsize;
	if (t->o % sizeof(SLOT)) t->o += sizeof(SLOT) - t->o % sizeof(SLOT);
	t->s    = t->o + sizeof(SLOT);

	t->m    = m = round((unsigned long)m);
	t->mask = m - 1; 
	t->l    = m * 2 / 3;

	t->n    = 0;

	e = (c = t->h = TMALLOC( char *, m  )) + m;
	while (c < e) *c++ = (char *)0;

	return t;
}

void hashgrow(Hashtable h)
{
	char ** o, ** p, ** n, ** N, ** ne;
	unsigned long newbit;

	h->h = TREALLOC( char *, h->h, h->m * 2  );

	/*
	 *  Rehash the entries, but keep them sorted.
	 */

	newbit   = h->m;

	for (N = (n = (o = h->h) + h->m) + h->m; n < N; n++, o++) {

		ne = n;	 /*  ne points to the current end of the new chain */ 
		p  = o;  /*  p  walks the old chain.			   */

		while (*p)
			if (INFO(h, *p)->hash & newbit) {

				* ne = * p;
				ne   = &NEXT(h, *ne);
				* p  = * ne;
			}
			else p = &NEXT(h, *p);

		*ne = (char *)0;
	}

	h->mask |= h->m;
	h->m    *= 2;
	h->l     = h->m * 2 / 3;
}

char * hash(Hashtable h, char const *str, int new)
{
	unsigned long i = hashf(str);
	char	   ** s = h->h + (i & h->mask);
	char	    * e;
	Slot	      slot;

	while (*s) {
		slot = INFO( h, *s );
		if (slot->hash >= i) {
			if (slot->hash > i) break;
			else if (!strcmp(STRING(h, *s), str)) return *s;
		}
		s = &slot->next;
	}

	if (!new) return (char *)0;

	e = TMALLOC(char, h->s + strlen(str) + 1);

	memset(e, '\0', h->o);			/* user data 	*/
	strcpy( STRING(h, e), str );		/* name		*/
	slot = INFO( h, e );			/* chaining	*/
	slot->hash = i;
	slot->next = *s;

	*s = e;

	h->n++;
	if (h->n >= h->l) hashgrow(h);
	return e;
}

char * hashnext(Hashtable h, char const * p) 
{
	if (p && NEXT(h, p)) return NEXT(h, p);
	else {
		register char ** s = h->h, ** e = s + h->m;

		if (p) s += (INFO(h, p)->hash & h->mask) + 1;
		while (s < e)
			if (*s++) return s[-1];
		return (char *)0;
	}
}

void hashdestroy(h) Hashtable h; {

	char ** s = h->h,
	     ** e = s + h->m;

	while (s < e) {
		char * n = *s++, * p;
		while (p = n) {
			n = NEXT(h, p);
			VFREE( p );
		}
	}
	VFREE( h->h );
	VFREE( h    );
}


char const * hashname(Hashtable h, char const * s)
{
	return STRING(h, s);
}


char const * intern(c) char const * c;
{
	static Hashtable h = (Hashtable)0;

	register void * v;

	if (!h) h = hashtable(0, 1000);
	v = hash( h, c, 1 );
	assert(v);

	return hashname(h, v);
}

int htst(h, c) Hashtable * h; char const * c;	/* test and set */
{
	register void * v;
	int		new_slot;

	assert(h);
	assert(c);

	if (!*h) *h = hashtable(0, 1000);
	if (new_slot = !(v = hash(*h, c, 0))) v = hash(*h, c, 1);
	assert(v);
	return !new_slot;
}
