#include "yacc2html.h"

extern char	      * program_name;

static unsigned char	in_attr;
static unsigned char	out_attr;
static char const     * url;

static unsigned int	newline_out = 1;

#define	PUTC(x)		((putchar(x) == EOF && (fprintf(stderr, 	\
	"%s: panic: putchar/write failed\n", program_name), exit(1), 0)))

#define PUTS(x)		(fputs((x), stdout) == EOF && (fprintf(stderr, \
	"%s: panic: fputs/write failed\n", program_name), exit(1), 0))

static char const * map[ /* uchar to HTML char or entity reference */ ] = {

	"&#0;", "&#1;", "&#2;", "&#3;", "&#4;", "&#5;", "&#6;", "&#7;", 
	"&#8;", "\t", "\n", "&#11;", "&#12;", "&#13;", "&#14;", "&#15;", 
	"&#16;", "&#17;", "&#18;", "&#19;", "&#20;", "&#21;", "&#22;", "&#23;", 
	"&#24;", "&#25;", "&#26;", "&#27;", "&#28;", "&#29;", "&#30;", "&#31;", 
	" ", "!", "&quot;", "#", "$", "%", "&amp;", "'", "(", ")", "*", "+",
	",", "-", ".", "/", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9",
	":", ";", "&lt;", "=", "&gt;", "?", "@", "A", "B", "C", "D", "E",
	"F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S",
	"T", "U", "V", "W", "X", "Y", "Z", "[", "\\", "]", "^", "_", "`", "a",
	"b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", 
	"p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z", "{", "|", "}",
	"~", "&#127;", "&#128;", "&#129;", "&#130;", "&#131;", "&#132;",
	"&#133;", "&#134;", "&#135;", "&#136;", "&#137;", "&#138;", "&#139;",
	"&#140;", "&#141;", "&#142;", "&#143;", "&#144;", "&#145;", "&#146;",
	"&#147;", "&#148;", "&#149;", "&#150;", "&#151;", "&#152;", "&#153;",
	"&#154;", "&#155;", "&#156;", "&#157;", "&#158;", "&#159;", "&#160;",
	"&#161;", "&#162;", "&#163;", "&#164;", "&#165;", "&#166;", "&#167;",
	"&#168;", "&#169;", "&#170;", "&#171;", "&#172;", "&#173;", "&#174;",
	"&#175;", "&#176;", "&#177;", "&#178;", "&#179;", "&#180;", "&#181;",
	"&#182;", "&#183;", "&#184;", "&#185;", "&#186;", "&#187;", "&#188;",
	"&#189;", "&#190;", "&#191;", "&#192;", "&#193;", "&#194;", "&#195;",
	"&#196;", "&#197;", "&#198;", "&#199;", "&#200;", "&#201;", "&#202;",
	"&#203;", "&#204;", "&#205;", "&#206;", "&#207;", "&#208;", "&#209;",
	"&#210;", "&#211;", "&#212;", "&#213;", "&#214;", "&#215;", "&#216;",
	"&#217;", "&#218;", "&#219;", "&#220;", "&#221;", "&#222;", "&#223;", 
	"&#224;", "&#225;", "&#226;", "&#227;", "&#228;", "&#229;", "&#230;",
	"&#231;", "&#232;", "&#233;", "&#234;", "&#235;", "&#236;", "&#237;",
	"&#238;", "&#239;", "&#240;", "&#241;", "&#242;", "&#243;", "&#244;",
	"&#245;", "&#246;", "&#247;", "&#248;", "&#249;", "&#250;", "&#251;",
	"&#252;", "&#253;", "&#254;", "&#255;"
};

static void OUTG(c, a, u) int c, a; char const * u;
{
	/*  At most one font attribute on the outside can be set at a time.
	 */
	assert(!((out_attr & A_FONT) & ((out_attr & A_FONT) - 1)));

	/* to do markup, turn off font and anchor attributes.
         */
	if (a & A_MARKUP) {

		if      (out_attr & A_BOLD      ) PUTS("</b>");
		else if (out_attr & A_ITALIC    ) PUTS("</i>");
		else if (out_attr & A_CODE      ) PUTS("</code>");
		else if (out_attr & A_TT	) PUTS("</tt>");
		else if (out_attr & A_EM        ) PUTS("</em>");

		if      (out_attr & A_ANCHOR    ) PUTS("</a>");

		out_attr &= ~(A_FONT|A_ANCHOR);

		newline_out = c == '\n';
		PUTC(c);
		return;
	}

	/* Figure out what it really looks like.
	 * Italic over bold; code over bold and italic; tt over c,b,i;
	 * em over everything.  (It's rare enough.)
	 */

	if (a & A_ITALIC) a &= ~A_BOLD;
	if (a & A_CODE  ) a &= ~(A_ITALIC|A_BOLD);
	if (a & A_TT    ) a &= ~(A_ITALIC|A_BOLD|A_CODE);
	if (a & A_EM    ) a &= ~(A_ITALIC|A_BOLD|A_CODE|A_TT);

	/*
	 *  You can't change attributes within an anchor. 
	 *  To do <a><b> </b><i> </i></a>, we have to do
	 *  <a><b> </b></a><a><i> </i></a>.
	 */

	if ((~a & A_FONT) & out_attr)
		a &= ~A_ANCHOR; 

	if ((out_attr & A_ANCHOR) != (a & A_ANCHOR)) {

		/* We are going to turn off or turn on anchor.
		   Turn off attributes before changing anchor. */

		if      (out_attr & A_BOLD      ) PUTS("</b>");
		else if (out_attr & A_ITALIC    ) PUTS("</i>");
		else if (out_attr & A_CODE      ) PUTS("</code>");
		else if (out_attr & A_TT	) PUTS("</tt>");
		else if (out_attr & A_EM        ) PUTS("</em>");
		out_attr &= ~A_FONT;

		if (a & A_ANCHOR) { PUTS("<a "); PUTS(u); PUTS(">"); }
		else PUTS("</a>");

		out_attr ^= A_ANCHOR;
	}

	if ((out_attr & A_FONT) != (a & A_FONT)) {

		if ((out_attr & A_BOLD  ) && !(a & A_BOLD  )) PUTS("</b>");
		if ((out_attr & A_ITALIC) && !(a & A_ITALIC)) PUTS("</i>");
		if ((out_attr & A_CODE  ) && !(a & A_CODE  )) PUTS("</code>");
		if ((out_attr & A_TT    ) && !(a & A_TT    )) PUTS("</tt>");
		if ((out_attr & A_EM    ) && !(a & A_EM    )) PUTS("</em>");

		if (!(out_attr & A_BOLD  ) && (a & A_BOLD  )) PUTS("<b>");
		if (!(out_attr & A_ITALIC) && (a & A_ITALIC)) PUTS("<i>");
		if (!(out_attr & A_CODE  ) && (a & A_CODE  )) PUTS("<code>");
		if (!(out_attr & A_TT    ) && (a & A_TT    )) PUTS("<tt>");
		if (!(out_attr & A_EM    ) && (a & A_EM    )) PUTS("<em>");

		out_attr = (out_attr & ~(A_FONT)) | (a & A_FONT);
	}

	/* Write the actual character, whew. */

	PUTS(  map[ (unsigned char)c ] );
	newline_out = c == '\n';
}

static void OUTC(i) int i;
{
	OUTG(i, in_attr, url);
}

static void SETA(i) int i;
{
	if (in_attr & i) {
		parser_error("SETA: attribute %x is already set\n", i);
		return;
	}
	in_attr |= i;
}

static void SETURL(c) char * c;
{
	url = intern(c);
	SETA(A_ANCHOR);
}

static int TSTA(i) int i;
{
	return !!(in_attr & i);
}

static void CLRA(i) int i;
{
	if (!(in_attr & i)) {
		parser_error("CLRA: attribute %d is not set\n", i);
		return;
	}
	in_attr &= ~i;
	if (i & A_ANCHOR) url = 0;
}


static void vprint(fmt, ap) char const * fmt; va_list ap;
{
	char const * original = fmt;
	static char buf	[1024];


	char	C;
	int	I;
	char	* S;

	for (; *fmt; fmt++) {
		if (*fmt != '%' || *++fmt == '%') OUTC(*fmt);
		else switch (*fmt) {
		case '<':
			switch (*++fmt) {
			case 'b': I = A_BOLD; 	goto seta;
			case 'i': I = A_ITALIC; goto seta;
			case 't': I = A_TT; 	goto seta;
			case 'e': I = A_EM; 	goto seta;
			case 'c': I = A_CODE; 	goto seta;
			case 'm': I = A_MARKUP;      seta:
				SETA(I);
				break;

			case 'a':	/* anchor	*/
				S = va_arg(ap, char *);
				SETURL(S);
				break;

			case 'n':	/* name		*/
			case 'H':	/* local href	*/
			case 'h':	/* href		*/
				S = va_arg(ap, char *);
				sprintf(buf,
				 *fmt=='h' ? "href=\"%s\"" :
				(*fmt=='H' ? "href=\"#%s\"" : "name=\"%s\""),
					S);
				SETURL(buf);
				break;

			default:
				parser_error(
			"vprint: bad character %c in \"%s\"\n",*fmt, original);
				return;
			}
			break;

		case '>':
			switch (*++fmt) {
			case 'b': I = A_BOLD; 	goto clra;
			case 'i': I = A_ITALIC; goto clra;
			case 't': I = A_TT; 	goto clra;
			case 'e': I = A_EM; 	goto clra;
			case 'c': I = A_CODE; 	goto clra;
			case 'a':
			case 'n':
			case 'h':
			case 'H': I = A_ANCHOR;	goto clra;
			case 'm': I = A_MARKUP;      clra:
				CLRA(I);
				break;
			default:
	parser_error("vprint: bad sequence >%c in \"%s\"\n", *fmt, original);
				return;
			}
			break;

		case '_':
			if (!TSTA(A_TT)) {
				OUTC('\240');
				SETA(A_TT);
				OUTC(' ');
				CLRA(A_TT);
			} else {
				OUTC(' ');
			}
			break;

		case 'm':
			S = va_arg(ap, char *);

			I = TSTA(A_MARKUP);
			if (!I) SETA(A_MARKUP);
			for (; *S; S++) OUTC(*S);
			if (!I) CLRA(A_MARKUP);

			break;

		case 'c':
			C = va_arg(ap, int);
			OUTC(C);
			break;

		case 'd':
			I = va_arg(ap, int);
			sprintf(buf, "%d", I);
			S = buf;
			goto string;

		case 's':
			S = va_arg(ap, char *);
		string: if (!S) S = "(null)";
			for (; *S; S++) OUTC(*S);
			break;

		default:
			fprintf(stderr, "vprint: panic: ill char %c\n", *fmt);
			exit(1);
		}
	}
	va_end(ap);
}

void print(char const * str, ...)
{
	va_list	ap;
	va_start(ap, str);

	vprint(str, ap);

	va_end(ap);
}

void  terminate_output()
{
	if (out_attr & A_BOLD  ) PUTS("</b>");
	if (out_attr & A_ITALIC) PUTS("</i>");
	if (out_attr & A_CODE  ) PUTS("</code>");
	if (out_attr & A_TT    ) PUTS("</tt>");
	if (out_attr & A_EM    ) PUTS("</em>");
	if (out_attr & A_ANCHOR) PUTS("</a>");

	if (!newline_out) {
		PUTC('\n');
		newline_out = 1;
	}
}
