#ifndef YACC2HTML_H
#define YACC2HTML_H

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

/* attributes for html output formatting */

#define	A_BOLD		2 
#define	A_ITALIC	4 
#define	A_TT		8 
#define	A_CODE		16 
#define	A_EM		32 
#define	A_MARKUP	64 	/* serious, "big" markup text	*/
#define	A_ANCHOR	128 	/* in anchor text		*/

#define	A_FONT		(A_BOLD | A_ITALIC | A_TT | A_CODE | A_EM)

/* memory */

#define	TMALLOC( type, nelems )		\
	((type *)Realloc( (char *)0, (size_t)(sizeof(type) * (nelems)) ))
#define	TREALLOC( type, ptr, nelems )	\
	((type *)(Realloc( (char *)(ptr), (size_t)(sizeof(type) * (nelems)) )))
#define	VFREE(ptr)		free(ptr)

char  	  * Realloc (char *, size_t);

/* parser input */

#define	INPUT_NONE		0
#define	INPUT_FILE		1
#define	INPUT_INTERACTIVE	2
#define	INPUT_REMOTE		3
#define	INPUT_N			4

extern struct ParserInputRec {
	int type;
	int (* f_input)();
	int (* f_unput)();
	union {
		struct {
			FILE * 	f;
			char *  name;
		} file;

		struct {
			unsigned char	lookahead;
			int 		prompt;
			FILE  	      * f;
		} interactive;
	} data;
} parser_input;

void parser_error(char const * fmt, ...);

void parser_set_interactive(FILE *);
void parser_set_file(char *, FILE *);
int parser_print_errors(void);

/* hashtable */

typedef struct _hashtable {

	unsigned long	mask;	/* mask for hashing, m-1		*/

	unsigned long	m;	/* # of slots, rounded to 2^n		*/
	unsigned long 	n;	/* # of currently occupied slots	*/
	unsigned long   l;	/* maximum fill before grow: 2/3 m  	*/

	unsigned long   s;	/* element size + sizeof(SLOT)		*/
	unsigned long   o;	/* slot offset: s - sizeof(SLOT)	*/

	char		** h;	/* actual hashtable			*/

} HASHTABLE, * Hashtable;

Hashtable	  hashtable(unsigned int, unsigned int);
void 		  hashdestroy(Hashtable);
char		* hash(Hashtable, char const *, int);
void		  hashdelete(Hashtable, char *);
char 		* hashnext(Hashtable, char const *);
char const	* hashname(Hashtable, char const *);
char const       * intern  (char const *);
int 	          htest(Hashtable *, char const *);

void		  define_link();

#define	htable( type, page )		hashtable( sizeof(type), page )
#define	hdestroy( table, type)		hashdestroy( table )
#define	hnew( table, type, str )	((type *)hash(table,str,1))
#define	haccess( table, type, str )	((type *)hash(table,str,0))
#define	hdelete( table, type, ob )	hashdelete(table, (char *)(ob))
#define	hname( table, type, ob )	hashname(table, (char *)(ob))
#define hnext(table, type, ob)		((type *)hashnext(table, (char *)(ob)))

/* misc */

char const       * uprint(char const *, char const *);
char const      ** definedP (char const * src);

/* output.c */ 

void print(char const * str, ...);
void terminate_output();

#endif /* YACC2HTML_H */
